#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
import logging
logger = logging.getLogger('PyAMS (SQLAlchemy)')

import sqlalchemy
import time
from datetime import datetime
from sqlalchemy.event import listens_for
from sqlalchemy.orm.scoping import scoped_session
from sqlalchemy.orm.session import sessionmaker
from sqlalchemy.pool import Pool, NullPool
from threading import Thread, Lock

# import interfaces
from pyams_alchemy.interfaces import REQUEST_SESSION_KEY, IAlchemyEngineUtility
from pyams_utils.interfaces.site import IOptionalUtility
from zope.lifecycleevent.interfaces import IObjectAddedEvent, IObjectModifiedEvent, IObjectRemovedEvent

# import packages
from persistent import Persistent
from persistent.dict import PersistentDict
from pyams_utils.property import DocFieldProperty
from pyams_utils.registry import query_utility
from pyams_utils.request import check_request, get_request_data, set_request_data
from pyams_utils.vocabulary import vocabulary_config
from pyramid.events import subscriber
from zope.componentvocabulary.vocabulary import UtilityVocabulary
from zope.container.contained import Contained
from zope.interface import implementer
from zope.sqlalchemy.datamanager import _SESSION_STATE, STATUS_READONLY, STATUS_ACTIVE, \
    ZopeTransactionExtension, join_transaction


CONNECTIONS_TIMESTAMP = {}
CONNECTIONS_LOCK = Lock()


@listens_for(Pool, 'checkout')
def handle_pool_checkout(connection, record, proxy):
    """Pool connection checkout

    Called when a connection is retrieved from the pool.
    If the connection record is already marked, we remove it from the mapping.
    """
    with CONNECTIONS_LOCK:
        if record in CONNECTIONS_TIMESTAMP:
            logger.debug("Removing timestamp for checked-out connection {0!r} ({1!r})".format(connection, record))
            del CONNECTIONS_TIMESTAMP[record]


@listens_for(Pool, 'checkin')
def handle_pool_checkin(connection, record):
    """Pool connection checkin

    Called when a connection returns to the pool.
    We apply a timestamp on the connection record to be able to close it automatically
    after 5 minutes without being used.
    """
    with CONNECTIONS_LOCK:
        logger.debug("Setting inactivity timestamp for checked-in connection {0!r} ({1!r})".format(connection, record))
        CONNECTIONS_TIMESTAMP[record] = datetime.utcnow()


class ConnectionCleanerThread(Thread):
    """Background thread used to clean unused database connections

    Each connection is referenced in CONNECTION_TIMESTAMPS mapping on checkin and is invalidated
    if not being used after 5 minutes
    """
    timeout = 300

    def run(self):
        while True:
            now = datetime.utcnow()
            for connection, value in list(CONNECTIONS_TIMESTAMP.items()):
                delta = now - value
                if delta.total_seconds() > self.timeout:
                    logger.debug("Invalidating unused connection {0!r} from pool".format(connection))
                    with CONNECTIONS_LOCK:
                        connection.invalidate()
                        del CONNECTIONS_TIMESTAMP[connection]
            time.sleep(60)

logger.info("Starting SQLAlchemy connections management thread")
cleaner_thread = ConnectionCleanerThread()
cleaner_thread.daemon = True
cleaner_thread.start()


@implementer(IAlchemyEngineUtility, IOptionalUtility)
class AlchemyEngineUtility(object):
    """SQLAlchemy engine utility"""

    name = DocFieldProperty(IAlchemyEngineUtility['name'])
    dsn = DocFieldProperty(IAlchemyEngineUtility['dsn'])
    echo = DocFieldProperty(IAlchemyEngineUtility['echo'])
    use_pool = DocFieldProperty(IAlchemyEngineUtility['use_pool'])
    pool_size = DocFieldProperty(IAlchemyEngineUtility['pool_size'])
    pool_recycle = DocFieldProperty(IAlchemyEngineUtility['pool_recycle'])
    echo_pool = DocFieldProperty(IAlchemyEngineUtility['echo_pool'])
    encoding = DocFieldProperty(IAlchemyEngineUtility['encoding'])
    convert_unicode = DocFieldProperty(IAlchemyEngineUtility['convert_unicode'])

    def __init__(self, name='', dsn='', echo=False, use_pool=True, pool_size=25, pool_recycle=-1, echo_pool=False,
                 encoding='utf-8', convert_unicode=False, **kwargs):
        self.name = name
        self.dsn = dsn
        self.echo = echo
        self.use_pool = use_pool
        self.pool_size = pool_size
        self.pool_recycle = pool_recycle
        self.echo_pool = echo_pool
        self.encoding = encoding
        self.convert_unicode = convert_unicode
        self.kw = PersistentDict()
        self.kw.update(kwargs)

    def __setattr__(self, key, value):
        super(AlchemyEngineUtility, self).__setattr__(key, value)
        if (key != '_v_engine') and hasattr(self, '_v_engine'):
            delattr(self, '_v_engine')

    def get_engine(self, use_pool=True):
        kw = {}
        kw.update(self.kw)
        if not (use_pool and self.use_pool):
            # Always create a new engine when pooling is disabled to help engine disposal
            return sqlalchemy.create_engine(self.dsn,
                                            echo=self.echo,
                                            poolclass=NullPool,
                                            encoding=self.encoding,
                                            convert_unicode=self.convert_unicode,
                                            strategy='threadlocal',
                                            **kw)
        else:
            # Store engine into volatile attributes when pooling is enabled
            engine = getattr(self, '_v_engine', None)
            if engine is None:
                engine = self._v_engine = sqlalchemy.create_engine(self.dsn,
                                                                   echo=self.echo,
                                                                   pool_size=self.pool_size,
                                                                   pool_recycle=self.pool_recycle,
                                                                   echo_pool=self.echo_pool,
                                                                   encoding=self.encoding,
                                                                   convert_unicode=self.convert_unicode,
                                                                   strategy='threadlocal',
                                                                   **kw)
            return engine

    def clear_engine(self):
        if hasattr(self, '_v_engine'):
            delattr(self, '_v_engine')


class PersistentAlchemyEngineUtility(Persistent, AlchemyEngineUtility, Contained):
    """Persistent implementation of SQLAlchemy engine utility"""


@subscriber(IObjectAddedEvent, context_selector=IAlchemyEngineUtility)
def handle_added_engine(event):
    """Register new SQLAlchemy engine when added"""
    manager = event.newParent
    manager.registerUtility(event.object, IAlchemyEngineUtility, name=event.object.name or '')


@subscriber(IObjectModifiedEvent, context_selector=IAlchemyEngineUtility)
def handle_modified_engine(event):
    """Clear SQLAlchemy engine volatile attributes when modified"""
    IAlchemyEngineUtility(event.object).clear_engine()


@subscriber(IObjectRemovedEvent, context_selector=IAlchemyEngineUtility)
def handle_removed_engine(event):
    """Un-register an SQLAlchemy engine when deleted"""
    manager = event.oldParent
    manager.unregisterUtility(event.object, IAlchemyEngineUtility, name=event.object.name or '')


def get_engine(engine, use_pool=True):
    """Get engine matching given utility name"""
    if isinstance(engine, str):
        engine = query_utility(IAlchemyEngineUtility, name=engine)
        if engine is not None:
            return engine.get_engine(use_pool)


def get_session(engine, join=True, status=STATUS_ACTIVE, request=None, alias=None,
                twophase=True, use_zope_extension=True, use_pool=True):
    """Get a new SQLALchemy session

    Session is stored in request and in session storage.
    See :func:`get_user_session` function to get arguments documentation.
    """
    if request is None:
        request = check_request()
    if not alias:
        alias = engine
    session_data = get_request_data(request, REQUEST_SESSION_KEY, {})
    session = session_data.get(alias)
    if session is None:
        _engine = get_engine(engine, use_pool)
        if use_zope_extension:
            factory = scoped_session(sessionmaker(bind=_engine,
                                                  twophase=twophase,
                                                  extension=ZopeTransactionExtension()))
        else:
            factory = sessionmaker(bind=_engine, twophase=twophase)
        session = factory()
        if join:
            join_transaction(session, initial_state=status)
        if status != STATUS_READONLY:
            _SESSION_STATE[id(session)] = session
        if session is not None:
            session_data[alias] = session
            set_request_data(request, REQUEST_SESSION_KEY, session_data)
    return session


def get_user_session(engine, join=True, status=STATUS_ACTIVE, request=None, alias=None,
                     twophase=True, use_zope_extension=True, use_pool=True):
    """Get a new SQLAlchemy session

    :param str engine: name of an SQLAlchemy engine session utility; if *engine* is not given as a string, it is
        returned as-is.
    :param bool join: if *True*, session is joined to the current Pyramid transaction
    :param str status: status of the new session; can be STATUS_ACTIVE or STATUS_READONLY
    :param request: currently running request
    :param str alias: alias to use in connections mapping for this session
    :param bool twophase: if *False*, session will be isolated and not included into two-phase transactions mechanism
    :param bool use_zope_extension: if *True*, use ZopeTransactionExtension scoped session
    :param bool use_pool: if *True*, this session will use a pool
    """
    if isinstance(engine, str):
        session = get_session(engine, join, status, request, alias, twophase, use_zope_extension, use_pool)
    else:
        session = engine
    return session


@vocabulary_config(name='PyAMS SQLAlchemy engines')
class EnginesVocabulary(UtilityVocabulary):
    """SQLAlchemy engines vocabulary"""

    interface = IAlchemyEngineUtility
    nameOnly = True
